"""initial

Revision ID: f29bac335186
Revises: 
Create Date: 2020-11-22 15:24:18.781971

"""
from alembic import op
import sqlalchemy as sa
from palaestrai.store.database_model import Model
from palaestrai.store.database_base import metadata


# revision identifiers, used by Alembic.
revision = "f29bac335186"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###
